% ******************************************************************************
% Prototypical AFEM example with optimally preconditioned GMRES method for a
% nonsymmetric linear elliptic second-order PDE.
% ******************************************************************************

%% parameters
nDofsMax = 1e6;
theta = 0.5;
pmax = 2;
lamalg = 0.1;

%% initialization for every polynomial degree
%[nElem, nDofs, errEst, iterations, time, directSolveTime] = deal(zeros(pmax, 1000));
for p = 2:pmax
    %% setup geometry & spaces
    printLogMessage('*** p = %d (of %d) ***', p, pmax)
    mesh = Mesh.loadFromGeometry('Lshape');
    mesh.refineUniform();
    fes = FeSpace(mesh, HigherOrderH1Fe(p), 'dirichlet', ':');
    u = FeFunction(fes);
    uEllExact = FeFunction(fes);
    u.setData(0);
    
    %% initialize output data structure
    %% create identifier
    identifier = num2str(p) + "_nDofsMax" + sprintf('%1g', nDofsMax) + ...
        "_theta" + num2str(fix(100*theta)) + ...
        "_lambda" + sprintf('%1g', lamalg);

    leveldata = LevelData();
    leveldata.metaData("identifier") = "S" + identifier;
    
    %% set problem data for -\Delta u = 1 on L-shape
    blf = BilinearForm();
    symmetricBlf = BilinearForm();
    lf = LinearForm();
    
    blf.a = Constant(mesh, 1);
    blf.b = Constant(mesh, [10; 10]);
    blf.c = Constant(mesh, 0);
    symmetricBlf.a = blf.a;
    
    lf.f = Constant(mesh, 1);
    
    %% set up solver and operator for nested iteration
    [solver, P] = IterativeSolver.choose(fes, symmetricBlf, "gmres", ...
        "multigrid");
    solver.tol = 1e-6;
    solver.maxIter = 50;
    
    %% adaptive loop
    ell = 0;
    meshSufficientlyFine = false;
    timepoint = timeStep(0);  % start time measurement
    runtime = 0;

    while ~meshSufficientlyFine
        %% setup
        %tic;
        ell = ell + 1;
        A = assemble(blf, fes);
        principalPart = assemble(symmetricBlf, fes);
        rhs = assemble(lf, fes);
        
        freeDofs = getFreeDofs(fes);
        A = A(freeDofs,freeDofs);
        principalPart = principalPart(freeDofs, freeDofs);
        solver.setupSystemMatrix(A, principalPart);
        solver.setupRhs(rhs(freeDofs), u.data(freeDofs)');
        
        solverIsConverged = false;
        k = 0;
        while ~solverIsConverged
            k = k + 1;
            solver.step();
            u.setFreeData(solver.x)
            eta2 = estimate(blf, lf, u);
            estimator = sqrt(sum(eta2));
            solver.tol = lamalg*estimator;
            solverIsConverged = solver.applyStoppingCriterion();
            % store level-oriented data
                leveldata.append('ell', int32(ell), 'k', int32(k), ...
                    'nDofs', int32(length(freeDofs)), ...
                    'nElem', int32(mesh.nElements), ...
                    'estimator', estimator, ...
                    'algebraicError', solver.residualCNorm,...
                    'case', 'A');

                if ~solverIsConverged
                    [timepoint, steptime] = timeStep(timepoint);
                    runtime = runtime + steptime;
                    leveldata.setTime(leveldata.nLevel, 'steptime', steptime,...
                        'runtime', runtime);
                    leveldata.printLevel();
                end
        end
        
        %% stoping criterion
        meshSufficientlyFine = (length(freeDofs) > nDofsMax);
        
        %% refine mesh
        if ~meshSufficientlyFine
            marked = markDoerflerBinning(eta2, theta);
            mesh.refineLocally(marked, 'NVB');
            u.setData(prolongate(P, u));
        end

        %% commandline output in case of mesh refinement
        leveldata.set(leveldata.nLevel, 'case', 'R');
        [timepoint, steptime] = timeStep(timepoint);
        runtime = runtime + steptime;
        leveldata.setTime(leveldata.nLevel, 'steptime', steptime,...
            'runtime', runtime);
        leveldata.printLevel();
    end
end


function indicators = estimate(blf, lf, u)

% abbreviate data structures
fes = u.fes;
p = fes.finiteElement.order;
mesh =  fes.mesh;
trafo = getAffineTransformation(mesh);

% compute volume residual element-wise for piecewise-constant coefficient a
res_handle = @(a, b, c, v, w, z, f) (a .* (v(1,:,:) + v(4,:,:)) - ...
    vectorProduct(b, w) - c .*  z + f).^2;
res = CompositeFunction(res_handle, blf.a, blf.b, blf.c, ...
    Hessian(u), Gradient(u), u, lf.f);
qr = QuadratureRule.ofOrder(2*p);
volumeResidual = integrateElement(res, qr);

% compute edge residual edge-wise
qr = QuadratureRule.ofOrder(max(p-1, 1), '1D');
f = CompositeFunction(@(p,a) a.*p, Gradient(u), blf.a);
dirichletBnd = getCombinedBndEdges(mesh, fes.bnd.dirichlet);
edgeResidual = integrateNormalJump(f, qr, ...
    @(j) zeros(size(j)), {}, dirichletBnd, ...    % no jump on dirichlet edges
    @(j) j.^2, {}, ':');                          % normal jump on inner edges

% combine the resdiuals suitably
hT = sqrt(trafo.area);
indicators = hT.^2 .* volumeResidual + ...
    hT .* sum(edgeResidual(mesh.element2edges), Dim.Vector);
end