% PgmresMG01 (abstract subclass of Iterative) Solves nonsymmetric linear equations
%   iteratively by an optimally preconditioned generalized minimal residual (GMRES) method.
%
% See also: IterativeSolver
%% NOTE: This is still a work in progress and not yet fully functional.

classdef PgmresMultigrid < IterativeSolver

    properties (SetAccess=protected, GetAccess=public)
        residualCNorm % norm of the preconditioned residual
    end

    properties (Dependent, Access=public)
        nLevels
    end
    
    properties (Access=protected)
        smoother
        projectedMatrix
        lhsNorm
    end

    properties (Access=protected)
        C % preconditioner
        residual % residual vector
        Cresidual % preconditioned residual vector
        searchDirection % search direction
        normb % norm of right-hand side
        H % Hessenberg matrix for Arnoldi process
        V % Krylov subspace basis
        PV % preconditioned Krylov subspace basis
        g % modified residual vector
        sine
        cosine
    end

    %% constructor
    methods
        function obj = PgmresMultigrid (smoother)
            arguments
                smoother (1,1)
            end

            obj@IterativeSolver();
            obj.smoother = smoother;
        end

        function n = get.nLevels(obj)
            n = obj.smoother.nLevels;
        end
     
    %% extend superclass methods
        function setupSystemMatrix(obj, A, symA, varargin)
            setupSystemMatrix@IterativeSolver(obj, A, symA);
            projectedA = obj.smoother.setup(symA, varargin{:});
            obj.projectedMatrix{obj.nLevels} = projectedA;

            % initialize Hessenberg matrix and Krylov vectors
            n = size(A, 1);
            obj.V = zeros(n, obj.maxIter + 1); % sparse matrix for Krylov vectors
            obj.PV = zeros(n, obj.maxIter + 1); % sparse matrix for preconditioned Krylov vectors
            obj.H = sparse(obj.maxIter + 1, obj.maxIter); % sparse matrix for Hessenberg matrix
        end

        function setupRhs(obj, b, varargin)
            setupRhs@IterativeSolver(obj, b, varargin{:});

            % initialize residual & search direction
            obj.b = b;
            obj.residual = b - obj.A * obj.x;
            [obj.Cresidual, ~] = obj.Vcycle(obj.residual);
            obj.searchDirection = obj.Cresidual;
            obj.residualCNorm = sqrt(dot(obj.residual, obj.Cresidual, 1));
            obj.normb = vecnorm(b, 2, 1);

            % initialize modified residual vector
            obj.g = zeros(obj.maxIter + 1, 1);
            obj.g(1) = obj.residualCNorm;

            obj.cosine = zeros(obj.maxIter-1, 1); % % cosine of Givens rotation
            obj.sine = zeros(obj.maxIter-1, 1); % % sine of Givens rotation
        end
    end

    %% implement abstract superclass methods
    methods (Access=public)
        function tf = isConverged(obj)
            tf = ((obj.iterationCount >= obj.maxIter) ...
                | (obj.residualCNorm < obj.tol) ...
                | (obj.residualCNorm ./ obj.normb < obj.tol));
        end
    end

    methods (Access=protected)
        function computeUpdate(obj)

            % compute first Krylov vector
            if obj.iterationCount == 0
                obj.V(:, 1) = obj.Cresidual / obj.residualCNorm; % normalize search direction
            end

            % current basis vector
            k = obj.iterationCount + 1; 
            AVk = obj.A * obj.V(:, k);

            obj.PV(:, 1) = obj.residual/obj.residualCNorm;

            % Arnoldi process
            for i = 1:k
                obj.H(i, k) = obj.V(:, i)' * AVk;
            end

            w = AVk - obj.PV(:, 1:k) * obj.H(1:k, k);
            %Cw = obj.C.apply(w);
            [Cw, ~] = obj.Vcycle(w);

            obj.H(k+1, k) = sqrt(Cw'*w); % next off-diagonal element

            % Apply Givens rotation
            for i = 1:k-1
                temp = obj.cosine(i) * obj.H(i, k) + obj.sine(i) * obj.H(i + 1, k);
                obj.H(i + 1, k) = -obj.sine(i) * obj.H(i, k) + obj.cosine(i) * obj.H(i + 1, k);
                obj.H(i, k) = temp;
            end

            beta = sqrt( obj.H(k, k)^2 + obj.H(k + 1, k)^2);
            obj.cosine(k) = obj.H(k, k) / beta;
            obj.sine(k) = obj.H(k + 1, k) / beta;
            obj.H(k, k) = beta;

            % Update residual vector
            obj.g(k + 1) = -obj.sine(k) * obj.g(k);
            obj.g(k) = obj.cosine(k) * obj.g(k);

            % solve upper triangular system
            y = zeros(k,1);
            y(k) = obj.g(k)/obj.H(k,k);
            for i=k-1:-1:1
                tmp = obj.H(i,i+1:k) * y(i+1:k);
                y(i) = (obj.g(i)-tmp) / obj.H(i,i);
            end
            obj.x = obj.x0 + obj.V(:, 1:k) * y;

            if(obj.H(k+1,k)==0)
                obj.residual = 0;
                return;
            end

            % update residual
            obj.residual = obj.b - obj.A * obj.x;
            [obj.Cresidual, ~] = obj.Vcycle(obj.residual);
            obj.residualCNorm = sqrt(dot(obj.residual, obj.Cresidual, 1));

            % set new search direction
            obj.V(:, k+1) = Cw / obj.H(k+1,k);
            obj.PV(:, k+1) = w / obj.H(k+1,k);
        end

        function [Cx, algEstimator] = Vcycle(obj, x)
            assert(isequal(size(x, 1), size(obj.A, 1)), ...
                'Setup for multilevel iteration not complete!')
            
            % if there is only one coarse level: exact solve
            L = obj.nLevels;
            if L == 1
                Cx = obj.A \ x;
                algEstimator = sqrt(dot(Cx, obj.A * Cx, Dim.Vector));
                return
            end

            % descending cascade: no smoothing
            res{L} = x;
            for k = L:-1:2
                res{k-1} = obj.smoother.restrict(res{k}, k);
            end

            % exact solve on coarsest level to compute accumulative lifting of x
            Cx = obj.projectedMatrix{1} \ res{1};
            eta2 = dot(Cx, obj.projectedMatrix{1} * Cx, Dim.Vector);

            % ascending cascade: (local) smoothing and compute error estimator
            for k = 2:(L-1)
                Cx = obj.smoother.prolongate(Cx, k);
                updatedResidual = res{k} - obj.projectedMatrix{k} * Cx;
                rho = obj.smoother.smooth(updatedResidual, k);
                [etaUpdate2, xUpdate] = computeOptimalUpdate(obj.projectedMatrix{k}, updatedResidual, rho);
                Cx = Cx + xUpdate;
                eta2 = eta2 + etaUpdate2;
            end

            % final smoothing and residual update (with non-projected matrix)
            Cx = obj.smoother.prolongate(Cx, L);
            updatedResidual = res{L} - obj.A * Cx;
            rho = obj.smoother.smooth(updatedResidual, L);
            [etaUpdate2, xUpdate] = computeOptimalUpdate(obj.A, updatedResidual, rho);
            Cx = Cx + xUpdate;
            algEstimator = sqrt(eta2 + etaUpdate2);
        end
    end
end


%% auxiliary functions
% error correction with optimal stepsize 
function [etaUpdate2, resUpdate] = computeOptimalUpdate(A, res, rho)
    rhoArho = dot(rho, A*rho, Dim.Vector);
    if max(abs(rho)) < eps
        lambda = 1; 
    else
        lambda = dot(res, rho, Dim.Vector) ./ rhoArho;
    end
    resUpdate = lambda .* rho;
    etaUpdate2 = rhoArho .* lambda.^2;
    
    if any(lambda > 3)
       warning('MG step-sizes no longer bound by d+1. Optimality of step size cannot be guaranteed!')
    end
end